import sys, os; sys.path.append(os.path.abspath(".."));
import src.qbandas as qb;
import pandas as pd

import src.qbandas.util as util

if __name__ == '__main__':
    
    # create a df for testing 
    df = pd.read_csv("data.csv")
    print("DATA: ", df)

    # test reading a schema from json
    schema = qb.read_schema('schema.json')
    print("SCHEMA: ")
    print("COL_TYPES: ", util.str_dict(schema[0]))
    print("FIDS: ", util.str_dict(schema[1]))

    # transform the data
    out = qb.format_values(df, schema[0])
    print("FORMATTED: ", out)

    out = qb.list_records(out, schema[1])
    print("RECORDS: ", util.str_dict(out))





import pandas as pd
from datetime import datetime

def _parse_default(x: object, args: tuple) -> dict:
    """ Pack a value into the default QuickBase API format.

    Parameters
    ----------
    x
        The value to pack. It can be anything.
    args
        unused. must be an empty tuple.

    Returns
    -------
    the packed value
    """

    if len(args):
        raise Exception(f"Default columns take no arguments. Cannot use args {args} for value {x} of type {type(x)}.")
    if pd.isna(x):
        return None
    return {'value':x}


def _parse_duration(x: float|int, args: tuple) -> dict:
    """ Pack a number into duration format for the QuickBase API
    
    Parameters
    ----------
    x
        The value to pack. It should be a number.
    args : tuple
        `agrs[0]` is the duration unit 'seconds' or 'milliseconds'. Defaults to 'seconds'

    Returns
    -------
    the packed value
    """
    if pd.isna(x):
        return None

    x = int(x)

    units = args[0] if len(args) else 'seconds'

    if units == 'seconds':
        x *= 1000
    elif units == 'milliseconds':
        pass
    else:
        raise Exception(f"'{args[0]}' is not a valid duration unit for value {x} of type {type(x)}")

    return {'value':x}
    
def _parse_date(x: datetime|str, args: tuple) -> dict:
    """ Pack a date into the date format for the QuickBase API

    Parameters
    ----------
    x
        The value to pack. It should be a datetime object or a string
    args : tuple
         `agrs[0]` is the parsing string 

    Returns
    -------
    the packed value
    """
    if pd.isna(x):
        return None

    format = args[0] if len(args) else "%m.%d.%Y"

    if not isinstance(x, (datetime, str)):
        raise Exception(f"'{type(x)}' cannot be coerced to date")
    elif isinstance(x, str):
        x = datetime.strptime(x, format)

    return {'value':x.strftime('%Y-%m-%d')}

def _parse_datetime(x: datetime|str, args: tuple) -> dict:
    """ Pack a datetime into the datetime format for the QuickBase API
    
    Parameters
    ----------
    x
        The value to pack. It should be a datetime object or a string
    args : tuple
         `agrs[0]` is the parsing string 

    Returns
    -------
    the packed value
    """
    if pd.isna(x):
        return None

    format = args[0] if len(args) else '%d%b%Y:%H:%M:%S.%f'

    if not isinstance(x, (datetime, str)):
        raise Exception(f"'{type(x)}' cannot be coerced to datetime")
    elif isinstance(x, str):
        x = datetime.strptime(x, format)
    
    return {'value':x.strftime('%Y-%m-%dT%H:%M:%SZ')}


def _parse_phonenum(x: None|str, args: tuple) -> dict:
    """ Pack a phone string into the phone format for the QuickBase API
    
    Parameters
    ----------
    x
        The value to pack. It should be a string
    args : tuple
        `agrs[0]` is the format string for reading the phone number. The phone number "(123) 456-7890 x123" would have the format string "(###) ###-#### x###". The extension must come last. It is optional for x to include it.

    Returns
    -------
    the packed value
    """
    format = args[0] if len(args) else "##########"
    
    if pd.isna(x):
        return None
    
    if not isinstance(x, str):
        raise Exception(f"type {type(x)} is invalid for value {x}, expected type str")

    # a str containing only the digits in order
    try:
        y = [x[i] for i in range(len(x)) if format[i] == '#']
        y = ''.join(y)
    except IndexError:
        raise Exception(f"could not parse '{x}' with format '{format}'") 

    if not y.isnumeric() or len(y) < 10:
        raise Exception(f"could not parse '{x}' with format '{format}'") 

    t = f'({y[0:3]}) {y[3:6]}-{y[6:10]}'
    if len(y) > 10:
        t = t + f' x{y[10:]}'

    #return "(123) 456-7890 x123"
    return {'value':t}


""" Integrates the popular data handling library Pandas and the QuickBase API
"""
import requests
from functools import partial
from typing import Iterator

import pandas as pd

from .config import field_types
from .schema import *

def format_values(df: pd.DataFrame, col_types: dict) -> pd.DataFrame:
    """ Reformats the values in a dataframe into values that the QuickBase API can use

    WARNING: Function not intended for enduser. See `qb.upload()`.
    
    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame containing the records that you want to send to QuickBase. 
    col_types : dict
        Controlls how the columns of the DataFrame are formatted. See `qb.read_schema()` for instructions on how to create and read in this dictonary. 

    Returns
    -------
    DataFrame : A copy of the original dataframe with it's values reformatted. 
    """

    # Figure out what columns are going where, set operations!
    requested_cols = set(col_types.keys())  # from col_types, user specced
    actual_cols = set(df.columns)           # from df

    # check for typos in the schema
    invalid_cols = requested_cols.difference(actual_cols)
    if len(invalid_cols):
        raise Exception(f"columns '{invalid_cols}' are invalid")

    # tell enduser about drops
    for col in actual_cols.difference(requested_cols):
        print(f'column \'{col}\' is being dropped')

    out = pd.DataFrame()

    # look at only the columns in schema and df
    for col in requested_cols.intersection(actual_cols):

        
        # parse the spec
        col_type, args = col_types[col][0], col_types[col][1:]
        print(f"processing column '{col}' as '{col_type}'")

        try:
            # Magic lines! Get the related parsing function from config, create
            # a partial of it with the given arguments, apply it to a copy of the
            # current column, and save it in the output
            parsing_func = field_types[col_type].parsing_func
            out[col] = df[col].copy().apply(partial(parsing_func, args=args))
        except Exception as e:
            raise Exception(f"Failed parsing column '{col}' with column type '{col_type}' and arguments '{args}'") from e

    return out 

def list_records(df: pd.DataFrame, fids: dict) -> list:
    """ Get a list of all the records in your dataframe.

    WARNING: Function not intended for enduser. See `qb.upload()`.
    
    Parameters
    ----------
    df : pandas.Dataframe
        The DataFrame containing the records that you want to send to QuickBase. The values in this DataFrame shoul be formatted for QuickBase. You can do this by calling `qb.format_values()` or write your own proceedure.
    fids : dict
        A mapping from columns in df to QuickBase field IDs as strings.  See `qb.read_schema()` for instructions on how to create and read in this dictonary. 
        
    Returns
    -------
    list : all the records in your table now ready to be sent to the QuickBase API.
    """
    
    # validate the fids
    for col in df.columns:
        if not col in fids.keys():
            raise Exception(f"column '{col}' not defined in fids")
        
    # output dataframe, remap cols to fids
    out = df.copy().rename(columns=fids)

    # Convert each row to a dictonary, drop keys with null values
    def _row_to_dict(row):
        "Convert a row into a dictonary"
        row = dict(row)                                         # makes dict
        return {k: v for k, v in row.items() if not pd.isna(v)} # drops null
    out = list(out.apply(_row_to_dict, axis=1)) 

    return out

def send_records(records: list, info: dict) -> Iterator[dict]:
    """ Send a list of records to a table in a QuickBase app. 

    WARNING: not intended for enduser. see `qb.upload()`.

    Parameters
    ----------
    records : list
        A list of dictonaries. Each dictonary is one record with field IDs for keys.
    info : dict
        Determines the which table in which QuickBase app to send to. See `qb.upload()` for more details. 

    Throws
    ------
    requests.HTTPError : on bad response from endpoint. 

    Returns
    -------
    Iterator[dict] : The json responses from QuickBase. This is mostly for advanced users who want to see stats and such. Visit [this link](https://developer.quickbase.com/operation/upsert) to see the formatting.
    """

    # setup destination information
    headers = {
        'QB-Realm-Hostname': info['QB-Realm-Hostname'],
        'User-Agent': info['User-Agent'] if 'User-Agent' in info else 'qBandas User',
        'Authorization': info['Authorization']
    }
    body = {"to": info['DBID']}

    # look for optinal arguments to include
    for option in ["fieldsToReturn", 'mergeFieldId']:
        if option in info:
            body[option] = info[option]

    # send the data one batch a time
    batch_size = info["Payload-Size"] if "Payload-Size" in info else 20_000
    for i in range(0, len(records), batch_size):

        body["data"] = records[i:i+batch_size]
        
        r = requests.post(
            'https://api.quickbase.com/v1/records', 
            headers = headers, 
            json = body
        )

        yield r
        

def upload(df: pd.DataFrame, schema: tuple|str, info: dict) -> Iterator[dict]:
    """
    Send tabular data to QuickBase.

    Parameters
    ----------
    df : pd.DataFrame
        The tabular data to send. 
    schema : tuple|str
        Specifies the field IDs and the data type (with additional formatting requirements) for each column in the dataframe. Use read_schema() to get this Python object. 
    info : dict
        Specify who is sending the data and where it is going. This info dict has three required keys. Here is an example `info` dict.
        
        ``` python
        info = {
            "QB-Realm-Hostname": 'demo.quickbase.com',
            "Authorization": 'QB-USER-TOKEN {TOKEN}',
            "DBID": "abcdef123"
        }
        ```

        Required arguments are "QB-Realm-Hostname", "Authorization", and "DBID". "DBID" is the destination table identifier on QuickBase. Can be found in the URL when viewing the table. You can also find it in the app protperties if you are an admin. "QB-Realm-Hostname" is your Quickbase subdomain. "Authorization" is the Quickbase authentication scheme you are using to authenticate the request, as described on the [authorization page](https://developer.quickbase.com/auth).

        Optional keys are "mergeFieldId", "User-Agent", and "fieldsToReturn". "mergeFieldId" gets a integer value to specify how to merge these new records into the exsiting records. The default is 3 or Record ID#. "fieldsToReturn" gets a list of integers specifying fields to get back after the upload. If you request any fields, you will automatically get the field's Records ID#s. "User-Agent" lets you name yourself for the API. Defaults to 'qBandas User'.

        A final optional key is "Payload-Size". This key takes and integer value that specifies the number of records to send per API request. The endpoint only supports payloads up to 10MB. Be careful with this setting, if it is too small performance will be significatly decreased; too large and you will have your requests rejected. Defaults to 20,000. 

        More information about all of these keys can be found on the QuickBase API [docs](https://developer.quickbase.com/operation/upsert). 

    Returns
    -------
    list[dict] : a list of responses from the endpoint about each payload you sent. Your df will be divided into payloads to accomodate the size limitations of the endpoint. 
    """
    col_types, fids = schema
    formatted = format_values(df, col_types)
    records = list_records(formatted, fids)
    return list(send_records(records, info))


    


import argparse, os, pathlib
import pandas as pd

from .schema import write_schema
from .config import field_types
from textwrap import fill

argParser = argparse.ArgumentParser(
    prog = 'qbandas',
    description = 'Utility functions to setup qBandas calls'
)

argParser.add_argument("-c", "--create-schema", 
    action='store_true',
    help='Create a schema file from the specified data file. See -d and -s.',
)

argParser.add_argument("-d", "--data-file", 
    type=pathlib.Path, 
    help="""Path to the data file. Defaults to 'data.csv' in current working directory. Only supports general delimited files (.csv, .tsv, etc.).""",
    metavar="file"
)

argParser.add_argument('-s', '--schema-file', 
    type=pathlib.Path, 
    help="Path to create schema file. Defaults to 'schema.json' in current working directory.",
    metavar="file"
)

argParser.add_argument('--delim',  
    help="The delimiter of the data file. Defaults to ','.",
    choices=[',', '\t', ' ', ';', '|'],
    default=',',
    metavar="delimiter"
)

argParser.add_argument("-t", "--col-types", 
    action='store_true',
    help='List the colmn types and their arguments.',
)

args = argParser.parse_args()
if args.create_schema:

    data_path = args.data_file
    if not data_path:
        data_path = os.path.join(os.getcwd(), 'data.csv')
    df = pd.read_table(data_path, delimiter=args.delim)

    schema_path = args.schema_file
    if not schema_path:
        schema_path = os.path.join(os.getcwd(), 'schema.json')
    write_schema(df, schema_path)
    
    exit()
    
if args.col_types:
    out = "data types:\n"
    for key in field_types:
        stub = f"  -{key}"
        gap = ' '*(16 - len(stub)) if len(stub) < 16 else '\n\t\t'
        out += stub + gap + fill(field_types[key], 80, subsequent_indent="\t\t") + '\n'
    print(out, end='')
    exit()



import shlex, json
import pandas as pd

def read_schema(path: str) -> tuple[dict, dict]:
    """ Read a schema from a JSON file and get a column types and field IDs dictionary

    Entries should have the format `"<column_name>": "<field_ID> <column_type> <*args>"`. This means each key is a column name, and each value is a field ID and column type plus additional formatting arguments. If you do not specify a column, it will be automatically dropped.
    
    Here is an example `schema.json`
    ```json
    {
        "PhoneNum": "7 phone ###.###.####",
        "FirstName": "8 text"
    }
    ```
    
    Parameters
    ----------
    path : str
        The location of the JSON file

    Returns
    -------
    tuple[dict, dict] : the first dictonary is the column types (plus any arguments) and the second is the field IDs
    """
    from .config import field_types
    col_types = {}
    fids = {}

    raw = json.load(open(path))
    
    for col in raw:
        value = shlex.split(raw[col])

        # check that everything is valid
        if len(value) < 2:
            raise Exception(f"column {col} has an invalid format {raw[col]}")
        if not str(value[0]).isdigit() or int(value[0]) < 1:
            raise Exception(f"column {col} has an invalid field ID {value[0]}")
        if not value[1] in field_types:
            raise Exception(f"column {col} has an invalid type {value[1]}")

        # save the arguments
        fids[col] = value[0]
        col_types[col] = value[1:]

    return col_types, fids

def write_schema(df: pd.DataFrame, path: str, space: int=32) -> None:
    """ Define a generic schema from a dataframe and write it to disk

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe to create a schema from
    path : str
        The path to the file to write the schema to
    space : int
        number of chars after the tab but before the type string
    """
    mapping = {
        'int64': 'numeric',
        'float64': 'numeric',
        'object': 'text',
        'str': 'text',
        'bool': 'checkbox',
        'datetime': 'datetime'
    }
    with open(path, 'a') as f:
    
        # check python's column types and get a list of the schema counterparts
        types = map(str, df.dtypes)
        types = map(lambda x: mapping[x] if x in mapping else None, types)
        
        # magic
        out = '{\n'
        for col, typ in zip(df.columns, types):
            b = ' '*(space-len(col)) if len(col) < 16 else ' '
            out += f'\t"{col}":' + b + f'"000 {typ}",\n'
        out = out[:-2] + '\n}'

        f.write(out)


import requests, json, re

def str_dict(x):
    """ Makes a readable string out of a dictionary
    """
    text = json.dumps(x, indent=2)
    text = re.sub(r"\{\s*(\"value\": .*)\s*\}", r"{\1}", text)
    text = re.sub(r"\},\s*\{", r"}, {", text)
    return text

def str_resp(r: requests.Response) -> str:
    """ Takes a resonse object and makes it into a string

    Parameters
    ----------
    r : requests.Response
        The response object with JSON

    Returns
    -------
    str : a string representation
    """

    text = str_dict(r.json())
    pattern = r"\[[\d,\s\n]+\]"
    matches = re.findall(pattern, text)
    for match in matches:
        n = match.count(',') 
        text = text.replace(match, f"[...{n+1} items...]")
    return text




from .parsers import _parse_default, _parse_duration, _parse_date, _parse_datetime, _parse_phonenum
from collections import namedtuple
Info = namedtuple('Info', ['info', 'parsing_func'])

field_types = {
    "text": Info(
        """string of text""",
        _parse_default),

    "rich-text": Info(
        """html text""",
        _parse_default),

    "multi-line-text": Info(
        """text with '\\n' support for newlines""",
        _parse_default),  

    "multiple-choice-text": Info(
        "string indicating selected option", 
        _parse_default),

    "multi-select": Info(
        "list of strings", 
        None),

    "address": Info(
        "composite field. there must be 4 columns in your datatframe for each address field on quickbase. street address, city, state, zip",
        _parse_default), 

    "email-address": Info(
        "valid email adresses as strings", 
        _parse_default),

    "url": Info(
        "string of a url", 
        _parse_default),

    "numeric": Info(
        "int or float data", 
        _parse_default),

    "numeric-percent": Info(
        "supports floats from zero to one inclusive",
        _parse_default), 

    "numeric-rating": Info(
        "integer from one to five", 
        _parse_default), 

    "numeric-currency": Info(
        "float indicating currency", 
        _parse_default),

    "duration": Info(
        "durration in milliseconds", 
        _parse_duration), 

    # eg. "2019-12-18"
    "date": Info(
        "date as a string or Datetime object. must specify date format as an argument eg. \"Y-D-M\" ", 
        _parse_date),

    # 'YYYY-MM-DDThh:mm:ssZ'
    "datetime": Info(
        "datetime as a string or a Datetime object. must specify datetime format as an argument eg. \"Y-D-M H:m:s\" ", 
        _parse_datetime),

    # military time hh:mm:ss.sss
    "time-of-day": Info(
        "military time as a string. must specify time-of-day format eg. \"hh:mm:ss.sss\"", 
        None),

    "checkbox": Info(
        "boolean as a int, string or bool. not case sensitve. false, 'no', 'n', 'f', 'false', 0 are all false, everything else is true except None", 
        _parse_default), 

    # "(123) 456-7890 x123"
    "phone-number": Info(
        "phone number as a string. must specify phone-number format eg. \"(###) ###-#### x####\" pound signs indicate digits. the first 10 digits are required, extension is optional.", 
        _parse_phonenum),

    # "value":
    "user": Info(
        "dictionary containing the user data that must include the user's ID eg. { \"id\": \"123456.ab1s\" }", 
        None),

    "list-user": Info(
        "a list of user dicts", 
        None),  # list of user dicts

    "file-attachment": Info(
        "Not supported", 
        None),

    "record-id": Info(
        "int. default merge field", 
        _parse_default),

    "lookup": Info(
        "Not supported", 
        None),

    "summary": Info(
        "Not supported", 
        None) 

}
